@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2022, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -16,6 +16,7 @@
 // Formulas at http://nines.cs.kuleuven.be/research/ecf/ecf.html
 
 #include "fem.hpp"
+#include "../mesh/nurbs.hpp"
 #include <cmath>
 
 #ifdef MFEM_USE_MPFR
@@ -35,6 +36,7 @@ IntegrationRule::IntegrationRule(IntegrationRule &irx, IntegrationRule &iry)
    ny = iry.GetNPoints();
    SetSize(nx * ny);
    SetPointIndices();
+   Order = std::min(irx.GetOrder(), iry.GetOrder());
 
    for (j = 0; j < ny; j++)
    {
@@ -59,6 +61,7 @@ IntegrationRule::IntegrationRule(IntegrationRule &irx, IntegrationRule &iry,
    const int nz = irz.GetNPoints();
    SetSize(nx*ny*nz);
    SetPointIndices();
+   Order = std::min({irx.GetOrder(), iry.GetOrder(), irz.GetOrder()});
 
    for (int iz = 0; iz < nz; ++iz)
    {
@@ -124,6 +127,7 @@ void IntegrationRule::GrundmannMollerSimplexRule(int s, int n)
    np /= f;
    SetSize(np);
    SetPointIndices();
+   Order = 2*s + 1;
 
    int pt = 0;
    for (int i = 0; i <= s; i++)
@@ -173,6 +177,52 @@ void IntegrationRule::GrundmannMollerSimplexRule(int s, int n)
    }
 }
 
+IntegrationRule*
+IntegrationRule::ApplyToKnotIntervals(KnotVector const& kv) const
+{
+   const int np = this->GetNPoints();
+   const int ne = kv.GetNE();
+
+   IntegrationRule *kvir = new IntegrationRule(ne * np);
+   kvir->SetOrder(GetOrder());
+
+   double x0 = kv[0];
+   double x1 = x0;
+
+   int id = 0;
+   for (int e=0; e<ne; ++e)
+   {
+      x0 = x1;
+
+      if (e == ne-1)
+      {
+         x1 = kv[kv.Size() - 1];
+      }
+      else
+      {
+         // Find the next unique knot
+         while (id < kv.Size() - 1)
+         {
+            id++;
+            if (kv[id] != x0)
+            {
+               x1 = kv[id];
+               break;
+            }
+         }
+      }
+
+      const double s = x1 - x0;
+
+      for (int j=0; j<this->GetNPoints(); ++j)
+      {
+         const double x = x0 + (s * (*this)[j].x);
+         (*kvir)[(e * np) + j].Set1w(x, (*this)[j].weight);
+      }
+   }
+
+   return kvir;
+}
 
 #ifdef MFEM_USE_MPFR
 
@@ -375,6 +425,7 @@ void QuadratureFunctions1D::GaussLegendre(const int np, IntegrationRule* ir)
 {
    ir->SetSize(np);
    ir->SetPointIndices();
+   ir->SetOrder(2*np - 1);
 
    switch (np)
    {
@@ -481,9 +532,11 @@ void QuadratureFunctions1D::GaussLobatto(const int np, IntegrationRule* ir)
    if ( np == 1 )
    {
       ir->IntPoint(0).Set1w(0.5, 1.0);
+      ir->SetOrder(1);
    }
    else
    {
+      ir->SetOrder(2*np - 3);
 
 #ifndef MFEM_USE_MPFR
 
@@ -578,6 +631,7 @@ void QuadratureFunctions1D::OpenUniform(const int np, IntegrationRule* ir)
 {
    ir->SetSize(np);
    ir->SetPointIndices();
+   ir->SetOrder(np - 1 + np%2);
 
    // The Newton-Cotes quadrature is based on weights that integrate exactly the
    // interpolatory polynomial through the equally spaced quadrature points.
@@ -594,6 +648,7 @@ void QuadratureFunctions1D::ClosedUniform(const int np,
 {
    ir->SetSize(np);
    ir->SetPointIndices();
+   ir->SetOrder(np - 1 + np%2);
    if ( np == 1 ) // allow this case as "closed"
    {
       ir->IntPoint(0).Set1w(0.5, 1.0);
@@ -612,6 +667,7 @@ void QuadratureFunctions1D::OpenHalfUniform(const int np, IntegrationRule* ir)
 {
    ir->SetSize(np);
    ir->SetPointIndices();
+   ir->SetOrder(np - 1 + np%2);
 
    // Open half points: the centers of np uniform intervals
    for (int i = 0; i < np ; ++i)
@@ -628,6 +684,7 @@ void QuadratureFunctions1D::ClosedGL(const int np, IntegrationRule* ir)
    ir->SetPointIndices();
    ir->IntPoint(0).x = 0.0;
    ir->IntPoint(np-1).x = 1.0;
+   ir->SetOrder(np - 1 + np%2); // Is this the correct order?
 
    if ( np > 2 )
    {
@@ -953,13 +1010,17 @@ const IntegrationRule &IntegrationRules::Get(int GeomType, int Order)
          if (!HaveIntRule(*ir_array, Order))
          {
             IntegrationRule *ir = GenerateIntegrationRule(GeomType, Order);
+#ifdef MFEM_DEBUG
             int RealOrder = Order;
             while (RealOrder+1 < ir_array->Size() &&
                    (*ir_array)[RealOrder+1] == ir)
             {
                RealOrder++;
             }
-            ir->SetOrder(RealOrder);
+            MFEM_VERIFY(RealOrder == ir->GetOrder(), "internal error");
+#else
+            MFEM_CONTRACT_VAR(ir);
+#endif
          }
       }
    }
@@ -1068,6 +1129,7 @@ IntegrationRule *IntegrationRules::PointIntegrationRule(int Order)
    IntegrationRule *ir = new IntegrationRule(1);
    ir->IntPoint(0).x = .0;
    ir->IntPoint(0).weight = 1.;
+   ir->SetOrder(1);
 
    PointIntRules[1] = PointIntRules[0] = ir;
 
@@ -1132,6 +1194,7 @@ IntegrationRule *IntegrationRules::SegmentIntegrationRule(int Order)
    {
       // Effectively passing memory management to SegmentIntegrationRules
       IntegrationRule *refined_ir = new IntegrationRule(2*n);
+      refined_ir->SetOrder(ir->GetOrder());
       for (int j = 0; j < n; j++)
       {
          refined_ir->IntPoint(j).x = ir->IntPoint(j).x/2.0;
@@ -1156,16 +1219,18 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
    // assuming that orders <= 25 are pre-allocated
    switch (Order)
    {
-      case 0:  // 1 point - 0 degree
+      case 0:  // 1 point - degree 1
       case 1:
          ir = new IntegrationRule(1);
          ir->AddTriMidPoint(0, 0.5);
+         ir->SetOrder(1);
          TriangleIntRules[0] = TriangleIntRules[1] = ir;
          return ir;
 
       case 2:  // 3 point - 2 degree
          ir = new IntegrationRule(3);
          ir->AddTriPoints3(0, 1./6., 1./6.);
+         ir->SetOrder(2);
          TriangleIntRules[2] = ir;
          // interior points
          return ir;
@@ -1174,6 +1239,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
          ir = new IntegrationRule(4);
          ir->AddTriMidPoint(0, -0.28125); // -9./32.
          ir->AddTriPoints3(1, 0.2, 25./96.);
+         ir->SetOrder(3);
          TriangleIntRules[3] = ir;
          return ir;
 
@@ -1181,6 +1247,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
          ir = new IntegrationRule(6);
          ir->AddTriPoints3(0, 0.091576213509770743460, 0.054975871827660933819);
          ir->AddTriPoints3(3, 0.44594849091596488632, 0.11169079483900573285);
+         ir->SetOrder(4);
          TriangleIntRules[4] = ir;
          return ir;
 
@@ -1189,6 +1256,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
          ir->AddTriMidPoint(0, 0.1125);
          ir->AddTriPoints3(1, 0.10128650732345633880, 0.062969590272413576298);
          ir->AddTriPoints3(4, 0.47014206410511508977, 0.066197076394253090369);
+         ir->SetOrder(5);
          TriangleIntRules[5] = ir;
          return ir;
 
@@ -1198,6 +1266,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
          ir->AddTriPoints3(3, 0.24928674517091042129, 0.058393137863189683013);
          ir->AddTriPoints6(6, 0.053145049844816947353, 0.31035245103378440542,
                            0.041425537809186787597);
+         ir->SetOrder(6);
          TriangleIntRules[6] = ir;
          return ir;
 
@@ -1212,6 +1281,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
                             0.30472650086816719592, 0.028775042784981585738);
          ir->AddTriPoints3R(9, 0.51584233435359177926, 0.27771616697639178257,
                             0.20644149867001643817, 0.067493187009802774463);
+         ir->SetOrder(7);
          TriangleIntRules[7] = ir;
          return ir;
 
@@ -1227,6 +1297,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
          ir->AddTriPoints6(10, 0.008394777409957605337213834539296,
                            0.263112829634638113421785786284643,
                            0.0136151570872174971324223450369544);
+         ir->SetOrder(8);
          TriangleIntRules[8] = ir;
          return ir;
 
@@ -1244,6 +1315,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
          ir->AddTriPoints6(13, 0.0368384120547362836348175987833851,
                            0.2219629891607656956751025276931919,
                            0.0216417696886446886446886446886446);
+         ir->SetOrder(9);
          TriangleIntRules[9] = ir;
          return ir;
 
@@ -1263,6 +1335,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
          ir->AddTriPoints6(19, 0.0095408154002994575801528096228873,
                            0.0668032510122002657735402127620247,
                            4.71083348186641172996373548344341E-03);
+         ir->SetOrder(10);
          TriangleIntRules[10] = ir;
          return ir;
 
@@ -1285,6 +1358,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
          ir->AddTriPoints6(22, 0.0448416775891304433090523914688007,
                            0.2772206675282791551488214673424523,
                            0.0205281577146442833208261574536469);
+         ir->SetOrder(11);
          TriangleIntRules[11] = ir;
          return ir;
 
@@ -1301,6 +1375,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
                            1.11783866011515E-02);
          ir->AddTriPoints6(27, 2.57340505483300E-02, 1.16251915907597E-01,
                            8.65811555432950E-03);
+         ir->SetOrder(12);
          TriangleIntRules[12] = ir;
          return ir;
 
@@ -1328,6 +1403,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
          ir->AddTriPoints6(31, 0.0897330604516053590796290561145196,
                            0.2723110556841851025078181617634414,
                            0.0182757511120486476280967518782978);
+         ir->SetOrder(13);
          TriangleIntRules[13] = ir;
          return ir;
 
@@ -1347,6 +1423,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
                            7.21815405676700E-03);
          ir->AddTriPoints6(36, 1.26833093287200E-03, 1.18974497696957E-01,
                            2.50511441925050E-03);
+         ir->SetOrder(14);
          TriangleIntRules[14] = ir;
          return ir;
 
@@ -1370,6 +1447,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
                            0.012803670460631195);
          ir->AddTriPoints6(48, 0.1684044181246992, 0.281835668099084562,
                            0.016544097765822835);
+         ir->SetOrder(15);
          TriangleIntRules[15] = ir;
          return ir;
 
@@ -1397,6 +1475,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
                             9.14639838501250E-03);
          ir->AddTriPoints6 (55, 1.46631822248280E-02, 8.07113136795640E-02,
                             3.33281600208250E-03);
+         ir->SetOrder(17);
          TriangleIntRules[16] = TriangleIntRules[17] = ir;
          return ir;
 
@@ -1428,6 +1507,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
                             0.0051292818680995);
          ir->AddTriPoints6 (67, 0.065494628082938, 0.010161119296278,
                             0.001899964427651);
+         ir->SetOrder(19);
          TriangleIntRules[18] = TriangleIntRules[19] = ir;
          return ir;
 
@@ -1462,6 +1542,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
                            0.009336472951467735);
          ir->AddTriPoints6(79, 0.140710844943938733,   0.323170566536257485,
                            0.01140911202919763);
+         ir->SetOrder(20);
          TriangleIntRules[20] = ir;
          return ir;
 
@@ -1513,6 +1594,7 @@ IntegrationRule *IntegrationRules::TriangleIntegrationRule(int Order)
                            0.00707722325261307);
          ir->AddTriPoints6(120, 0.191771865867325067,   0.325618122595983752,
                            0.007440689780584005);
+         ir->SetOrder(25);
          TriangleIntRules[21] =
             TriangleIntRules[22] =
                TriangleIntRules[23] =
@@ -1563,6 +1645,7 @@ IntegrationRule *IntegrationRules::TetrahedronIntegrationRule(int Order)
       case 1:
          ir = new IntegrationRule(1);
          ir->AddTetMidPoint(0, 1./6.);
+         ir->SetOrder(1);
          TetrahedronIntRules[0] = TetrahedronIntRules[1] = ir;
          return ir;
 
@@ -1570,6 +1653,7 @@ IntegrationRule *IntegrationRules::TetrahedronIntegrationRule(int Order)
          ir = new IntegrationRule(4);
          // ir->AddTetPoints4(0, 0.13819660112501051518, 1./24.);
          ir->AddTetPoints4b(0, 0.58541019662496845446, 1./24.);
+         ir->SetOrder(2);
          TetrahedronIntRules[2] = ir;
          return ir;
 
@@ -1577,6 +1661,7 @@ IntegrationRule *IntegrationRules::TetrahedronIntegrationRule(int Order)
          ir = new IntegrationRule(5);
          ir->AddTetMidPoint(0, -2./15.);
          ir->AddTetPoints4b(1, 0.5, 0.075);
+         ir->SetOrder(3);
          TetrahedronIntRules[3] = ir;
          return ir;
 
@@ -1585,6 +1670,7 @@ IntegrationRule *IntegrationRules::TetrahedronIntegrationRule(int Order)
          ir->AddTetPoints4(0, 1./14., 343./45000.);
          ir->AddTetMidPoint(4, -74./5625.);
          ir->AddTetPoints6(5, 0.10059642383320079500, 28./1125.);
+         ir->SetOrder(4);
          TetrahedronIntRules[4] = ir;
          return ir;
 
@@ -1595,6 +1681,7 @@ IntegrationRule *IntegrationRules::TetrahedronIntegrationRule(int Order)
          ir->AddTetPoints4(6, 0.092735250310891226402, 0.012248840519393658257);
          ir->AddTetPoints4b(10, 0.067342242210098170608,
                             0.018781320953002641800);
+         ir->SetOrder(5);
          TetrahedronIntRules[5] = ir;
          return ir;
 
@@ -1608,6 +1695,7 @@ IntegrationRule *IntegrationRules::TetrahedronIntegrationRule(int Order)
                             9.2261969239424536825E-03);
          ir->AddTetPoints12(12, 0.063661001875017525299, 0.26967233145831580803,
                             8.0357142857142857143E-03);
+         ir->SetOrder(6);
          TetrahedronIntRules[6] = ir;
          return ir;
 
@@ -1621,6 +1709,7 @@ IntegrationRule *IntegrationRules::TetrahedronIntegrationRule(int Order)
          ir->AddTetPoints4b(15, 2.3825066607381275412E-03,
                             4.8914252630734993858E-03);
          ir->AddTetPoints12(19, 0.1, 0.2, 0.027557319223985890653);
+         ir->SetOrder(7);
          TetrahedronIntRules[7] = ir;
          return ir;
 
@@ -1638,6 +1727,7 @@ IntegrationRule *IntegrationRules::TetrahedronIntegrationRule(int Order)
                             5.7044858086819185068E-03);
          ir->AddTetPoints4(38, 0.20682993161067320408, 0.014250305822866901248);
          ir->AddTetMidPoint(42, -0.020500188658639915841);
+         ir->SetOrder(8);
          TetrahedronIntRules[8] = ir;
          return ir;
 
@@ -1668,6 +1758,7 @@ IntegrationRule *IntegrationRules::PyramidIntegrationRule(int Order)
    int npts = irc.GetNPoints();
    AllocIntRule(PyramidIntRules, Order);
    PyramidIntRules[Order] = new IntegrationRule(npts);
+   PyramidIntRules[Order]->SetOrder(Order); // FIXME: see comment above
 
    for (int k=0; k<npts; k++)
    {
@@ -1690,6 +1781,7 @@ IntegrationRule *IntegrationRules::PrismIntegrationRule(int Order)
    int ns = irs.GetNPoints();
    AllocIntRule(PrismIntRules, Order);
    PrismIntRules[Order] = new IntegrationRule(nt * ns);
+   PrismIntRules[Order]->SetOrder(std::min(irt.GetOrder(), irs.GetOrder()));
 
    for (int ks=0; ks<ns; ks++)
    {
@@ -1725,4 +1817,235 @@ IntegrationRule *IntegrationRules::CubeIntegrationRule(int Order)
    return CubeIntRules[Order];
 }
 
+IntegrationRule& NURBSMeshRules::GetElementRule(const int elem,
+                                                const int patch, const int *ijk,
+                                                Array<const KnotVector*> const& kv,
+                                                bool & deleteRule) const
+{
+   deleteRule = false;
+
+   // First check whether a rule has been assigned to element index elem.
+   auto search = elementToRule.find(elem);
+   if (search != elementToRule.end())
+   {
+      return *elementRule[search->second];
+   }
+
+   MFEM_VERIFY(patchRules1D.NumRows(),
+               "Undefined rule in NURBSMeshRules::GetElementRule");
+
+   // Use a tensor product of rules on the patch.
+   MFEM_VERIFY(kv.Size() == dim, "");
+
+   int np = 1;
+   std::vector<std::vector<double>> el(dim);
+
+   std::vector<int> npd;
+   npd.assign(3, 0);
+
+   for (int d=0; d<dim; ++d)
+   {
+      const int order = kv[d]->GetOrder();
+
+      const double kv0 = (*kv[d])[order + ijk[d]];
+      const double kv1 = (*kv[d])[order + ijk[d] + 1];
+
+      const bool rightEnd = (order + ijk[d] + 1) == (kv[d]->Size() - 1);
+
+      for (int i=0; i<patchRules1D(patch,d)->Size(); ++i)
+      {
+         const IntegrationPoint& ip = (*patchRules1D(patch,d))[i];
+         if (kv0 <= ip.x && (ip.x < kv1 || rightEnd))
+         {
+            const double x = (ip.x - kv0) / (kv1 - kv0);
+            el[d].push_back(x);
+            el[d].push_back(ip.weight);
+         }
+      }
+
+      npd[d] = el[d].size() / 2;
+      np *= npd[d];
+   }
+
+   IntegrationRule *irp = new IntegrationRule(np);
+   deleteRule = true;
+
+   // Set (*irp)[i + j*npd[0] + k*npd[0]*npd[1]] =
+   //     (el[0][2*i], el[1][2*j], el[2][2*k])
+
+   MFEM_VERIFY(npd[0] > 0 && npd[1] > 0, "Assuming 2D or 3D");
+
+   for (int i = 0; i < npd[0]; ++i)
+   {
+      for (int j = 0; j < npd[1]; ++j)
+      {
+         for (int k = 0; k < std::max(npd[2], 1); ++k)
+         {
+            const int id = i + j*npd[0] + k*npd[0]*npd[1];
+            (*irp)[id].x = el[0][2*i];
+            (*irp)[id].y = el[1][2*j];
+
+            (*irp)[id].weight = el[0][(2*i)+1];
+            (*irp)[id].weight *= el[1][(2*j)+1];
+
+            if (npd[2] > 0)
+            {
+               (*irp)[id].z = el[2][2*k];
+               (*irp)[id].weight *= el[2][(2*k)+1];
+            }
+         }
+      }
+   }
+
+   return *irp;
+}
+
+void NURBSMeshRules::GetIntegrationPointFrom1D(const int patch, int i, int j,
+                                               int k, IntegrationPoint & ip)
+{
+   MFEM_VERIFY(patchRules1D.NumRows() > 0,
+               "Assuming patchRules1D is set.");
+
+   ip.weight = (*patchRules1D(patch,0))[i].weight;
+   ip.x = (*patchRules1D(patch,0))[i].x;
+
+   if (dim > 1)
+   {
+      ip.weight *= (*patchRules1D(patch,1))[j].weight;
+      ip.y = (*patchRules1D(patch,1))[j].x;  // 1D rule only has x
+   }
+
+   if (dim > 2)
+   {
+      ip.weight *= (*patchRules1D(patch,2))[k].weight;
+      ip.z = (*patchRules1D(patch,2))[k].x;  // 1D rule only has x
+   }
+}
+
+void NURBSMeshRules::Finalize(Mesh const& mesh)
+{
+   if ((int) pointToElem.size() == npatches) { return; }  // Already set
+
+   MFEM_VERIFY(elementToRule.empty() && patchRules1D.NumRows() > 0
+               && npatches > 0, "Assuming patchRules1D is set.");
+   MFEM_VERIFY(mesh.NURBSext, "");
+   MFEM_VERIFY(mesh.Dimension() == dim, "");
+
+   pointToElem.resize(npatches);
+   patchRules1D_KnotSpan.resize(npatches);
+
+   // First, find all the elements in each patch.
+   std::vector<std::vector<int>> patchElements(npatches);
+
+   for (int e=0; e<mesh.GetNE(); ++e)
+   {
+      patchElements[mesh.NURBSext->GetElementPatch(e)].push_back(e);
+   }
+
+   Array<int> ijk(3);
+   Array<int> maxijk(3);
+   Array<int> np(3);  // Number of points in each dimension
+   ijk = 0;
+
+   Array<const KnotVector*> pkv;
+
+   for (int p=0; p<npatches; ++p)
+   {
+      patchRules1D_KnotSpan[p].resize(dim);
+
+      // For each patch, get the range of ijk.
+      mesh.NURBSext->GetPatchKnotVectors(p, pkv);
+      MFEM_VERIFY((int) pkv.Size() == dim, "");
+
+      maxijk = 1;
+      np = 1;
+      for (int d=0; d<dim; ++d)
+      {
+         maxijk[d] = pkv[d]->GetNKS();
+         np[d] = patchRules1D(p,d)->Size();
+      }
+
+      // For each patch, set a map from ijk to element index.
+      Array3D<int> ijk2elem(maxijk[0], maxijk[1], maxijk[2]);
+      ijk2elem = -1;
+
+      for (auto elem : patchElements[p])
+      {
+         mesh.NURBSext->GetElementIJK(elem, ijk);
+         MFEM_VERIFY(ijk2elem(ijk[0], ijk[1], ijk[2]) == -1, "");
+         ijk2elem(ijk[0], ijk[1], ijk[2]) = elem;
+      }
+
+      // For each point, find its ijk and from that its element index.
+      // It is assumed here that the NURBSFiniteElement kv the same as the
+      // patch kv.
+
+      for (int d=0; d<dim; ++d)
+      {
+         patchRules1D_KnotSpan[p][d].SetSize(patchRules1D(p,d)->Size());
+
+         for (int r=0; r<patchRules1D(p,d)->Size(); ++r)
+         {
+            const IntegrationPoint& ip = (*patchRules1D(p,d))[r];
+
+            const int order = pkv[d]->GetOrder();
+
+            // Find ijk_d such that ip.x is in the corresponding knot-span.
+            int ijk_d = 0;
+            bool found = false;
+            while (!found)
+            {
+               const double kv0 = (*pkv[d])[order + ijk_d];
+               const double kv1 = (*pkv[d])[order + ijk_d + 1];
+
+               const bool rightEnd = (order + ijk_d + 1) == (pkv[d]->Size() - 1);
+
+               if (kv0 <= ip.x && (ip.x < kv1 || rightEnd))
+               {
+                  found = true;
+               }
+               else
+               {
+                  ijk_d++;
+               }
+            }
+
+            patchRules1D_KnotSpan[p][d][r] = ijk_d;
+         }
+      }
+
+      pointToElem[p].SetSize(np[0], np[1], np[2]);
+      for (int i=0; i<np[0]; ++i)
+         for (int j=0; j<np[1]; ++j)
+            for (int k=0; k<np[2]; ++k)
+            {
+               const int elem = ijk2elem(patchRules1D_KnotSpan[p][0][i],
+                                         patchRules1D_KnotSpan[p][1][j],
+                                         patchRules1D_KnotSpan[p][2][k]);
+               MFEM_VERIFY(elem >= 0, "");
+               pointToElem[p](i,j,k) = elem;
+            }
+   } // Loop (p) over patches
+}
+
+void NURBSMeshRules::SetPatchRules1D(const int patch,
+                                     std::vector<const IntegrationRule*> & ir1D)
+{
+   MFEM_VERIFY((int) ir1D.size() == dim, "Wrong dimension");
+
+   for (int i=0; i<dim; ++i)
+   {
+      patchRules1D(patch,i) = ir1D[i];
+   }
+}
+
+NURBSMeshRules::~NURBSMeshRules()
+{
+   for (int i=0; i<patchRules1D.NumRows(); ++i)
+      for (int j=0; j<patchRules1D.NumCols(); ++j)
+      {
+         delete patchRules1D(i, j);
+      }
+}
+
 }
